-- Preparation:

USE Playground;

-- Demo tables for some of the demos
DROP TABLE IF EXISTS dbo.Fruits;
DROP TABLE IF EXISTS dbo.Colors;

CREATE TABLE dbo.Fruits
    (FruitNum  int         NOT NULL,
     FruitName varchar(35) NOT NULL);
CREATE TABLE dbo.Colors
    (ColorNum  int         NOT NULL,
     ColorName varchar(35) NOT NULL,
     ColorCode binary(3)   NOT NULL);

CREATE CLUSTERED INDEX cix_Fruits_FruitNum
ON dbo.Fruits(FruitNum);
CREATE NONCLUSTERED INDEX cix_Fruits_FruitName
ON dbo.Fruits(FruitName);

CREATE CLUSTERED INDEX cix_Colors_ColorNum
ON dbo.Colors(ColorNum);
CREATE NONCLUSTERED INDEX cix_Colors_ColorName
ON dbo.Colors(ColorName);

INSERT INTO dbo.Fruits (FruitNum, FruitName)
VALUES (1, 'Apple'),
       (3, 'Cherry'),
       (2, 'Lime'),
       (3, 'Melon'),
       (2, 'Orange'),
       (5, 'Banana');

INSERT INTO dbo.Colors (ColorNum, ColorName, ColorCode)
VALUES (4, 'Blue',   0x0000FF),
       (2, 'Orange', 0xFFA500),
       (1, 'Red',    0xFF0000),
       (2, 'Yellow', 0xFFFF00),
       (5, 'Green',  0x008000),
       (6, 'Cyan',   0x00FFFF),
       (6, 'Teal',   0x008080);
GO

---------

-- Sample data for prefetching and optimized nested loops
-- Adapted from: https://learn.microsoft.com/en-us/archive/blogs/craigfr/optimizing-io-performance-by-sorting-part-1
-- Note: This script may take a few minutes to complete!

-- Minimum required database size: 4GB data and 1GB log
USE Playground;
SET NOCOUNT ON;

DROP TABLE IF EXISTS dbo.T;
CREATE TABLE dbo.T
    (PK      int       NOT NULL IDENTITY PRIMARY KEY,
     RandKey int       NOT NULL,
     Flags   tinyint   NOT NULL,
     Data    int       NOT NULL,
     Pad     char(100) NOT NULL);
DECLARE @I int = 0;
WHILE @I < 100000
  BEGIN;
    WITH
      X2 (R) AS ( SELECT RAND() UNION ALL SELECT RAND() ),
      X4 (R) AS ( SELECT R FROM X2 UNION ALL SELECT R FROM X2 ),
      X8 (R) AS ( SELECT R FROM X4 UNION ALL SELECT R FROM X4 ),
      X16 (R) AS ( SELECT R FROM X8 UNION ALL SELECT R FROM X8 ),
      X32 (R) AS ( SELECT R FROM X16 UNION ALL SELECT R FROM X16 ),
      X64 (R) AS ( SELECT R FROM X32 UNION ALL SELECT R FROM X32 ),
      X128 (R) AS ( SELECT R FROM X64 UNION ALL SELECT R FROM X64 ),
      X256 (R) AS ( SELECT R FROM X128 UNION ALL SELECT R FROM X128 )
    INSERT T (RandKey, Flags, Data, Pad)
    SELECT R * 1000000000, 0xFF, 1, ''
	FROM   X256;
    SET @I += 1;
  END;
CREATE INDEX IRandKey on T (RandKey, Flags);


-- Demo 1:

USE Playground;

SELECT     f.FruitNum,
           f.FruitName,
           c.ColorName
FROM       dbo.Fruits AS f
INNER JOIN dbo.Colors AS c
   ON      c.ColorNum = f.FruitNum;

SELECT          f.FruitNum,
                f.FruitName,
                c.ColorName
FROM            dbo.Fruits AS f
LEFT OUTER JOIN dbo.Colors AS c
  ON            c.ColorNum = f.FruitNum;


-- Demo 2:

USE AdventureWorks;

SELECT     soh.SalesOrderID,
           soh.RevisionNumber,
           soh.ModifiedDate
FROM       Sales.SalesOrderHeader AS soh
INNER JOIN Sales.SalesOrderDetail AS sod
   ON      sod.SalesOrderID = soh.SalesOrderID
OPTION (LOOP JOIN, MAXDOP 1);

SELECT     DISTINCT soh.SalesOrderID,
                    soh.RevisionNumber,
                    soh.ModifiedDate
FROM       Sales.SalesOrderHeader AS soh
INNER JOIN Sales.SalesOrderDetail AS sod
   ON      sod.SalesOrderID = soh.SalesOrderID
OPTION (LOOP JOIN, MAXDOP 1);


-- Demo 3:

USE Playground;

SELECT          f.FruitNum,
                f.FruitName,
                c.ColorName
FROM            dbo.Fruits AS f
FULL OUTER JOIN dbo.Colors AS c
  ON            f.FruitNum = c.ColorNum;

-- Second window

--SET STATISTICS IO ON;
SELECT f.FruitNum,
       f.FruitName,
       c.ColorNames
FROM   dbo.Fruits          AS f
FULL OUTER JOIN
       (SELECT   ColorNum,
                 STRING_AGG (ColorName, '/') AS ColorNames
        FROM     dbo.Colors
        GROUP BY ColorNum) AS c
  ON f.FruitNum > c.ColorNum
OPTION (LOOP JOIN);


-- Demo 4:

USE Playground;

SELECT     f.FruitName,
           c.ColorName
FROM       dbo.Fruits AS f
INNER JOIN dbo.Colors AS c
   ON      f.FruitNum = c.ColorNum
--ORDER BY   f.FruitName;

SELECT          f.FruitName,
                c.ColorName
FROM            dbo.Fruits AS f
LEFT OUTER JOIN dbo.Colors AS c
  ON            f.FruitNum = c.ColorNum
--ORDER BY        f.FruitName;

SELECT   f.FruitName
FROM     dbo.Fruits AS f
WHERE    EXISTS
    (SELECT * FROM dbo.Colors AS c WHERE f.FruitNum = c.ColorNum)
--ORDER BY f.FruitName;

SELECT   f.FruitName
FROM     dbo.Fruits AS f
WHERE    NOT EXISTS
    (SELECT * FROM dbo.Colors AS c WHERE f.FruitNum = c.ColorNum)
--ORDER BY f.FruitName;

SELECT   f.FruitName,
         CASE WHEN EXISTS
                       (SELECT * FROM dbo.Colors AS c WHERE f.FruitNum = c.ColorNum)
                  THEN 'Yes'
              ELSE 'No'
         END AS ColorExists
FROM     dbo.Fruits AS f
--ORDER BY f.FruitName;


-- Demo 5:

USE Playground;

--SET STATISTICS IO, TIME ON;
DECLARE @Boundary int = 2500000;

SELECT   PK,
         RandKey,
         Data
FROM     dbo.T
WHERE    RandKey < @Boundary
ORDER BY RandKey
OPTION (MAXDOP 1, OPTIMIZE FOR (@Boundary = 500000));

-- Second window:

-- First check your "normal" maximum server memory setting
EXEC sp_configure 'max server memory';
-- Copy config_value and paste it in the commented query below

-- Now force a ridiculously low maximum server memory
EXEC sp_configure 'max server memory', 1024;
RECONFIGURE WITH OVERRIDE;

-- When done with the demos, uncomment and run this to return to normal
--EXEC sp_configure 'max server memory', 23424; -- Paste your "normal" value here!
--RECONFIGURE WITH OVERRIDE;

-- Third window:

SELECT   node_id,
         thread_id,
         physical_operator_name,
         row_count
FROM     sys.dm_exec_query_profiles
WHERE    session_id = 88 -- Replace with spid of first window (with SELECT query).
ORDER BY node_id;


-- Demo 6:

USE Playground;

--SET STATISTICS IO, TIME ON;
DECLARE @Boundary int = 2500000;

SELECT PK,
       RandKey,
       Data
FROM   dbo.T
WHERE  RandKey < @Boundary
--ORDER BY RandKey
OPTION (MAXDOP 1, OPTIMIZE FOR (@Boundary = 500000),
        USE HINT ('DISABLE_OPTIMIZED_NESTED_LOOP'));

-- Second window:

-- First check your "normal" maximum server memory setting
EXEC sp_configure 'max server memory';
-- Copy config_value and paste it in the commented query below

-- Now force a ridiculously low maximum server memory
EXEC sp_configure 'max server memory', 1024;
RECONFIGURE WITH OVERRIDE;

-- When done with the demos, uncomment and run this to return to normal
--EXEC sp_configure 'max server memory', 23424; -- Paste your "normal" value here!
--RECONFIGURE WITH OVERRIDE;


-- Demo 7:

USE Playground;

--SET STATISTICS IO, TIME ON;
DECLARE @Boundary int = 2500000;

SELECT PK,
       RandKey,
       Data
FROM   dbo.T
WHERE  RandKey < @Boundary
--ORDER BY RandKey
OPTION (MAXDOP 1, OPTIMIZE FOR (@Boundary = 500000),
        USE HINT ('DISABLE_OPTIMIZED_NESTED_LOOP'));


-- Cleanup:

-- When done with the demos, uncomment and run this to return to normal
EXEC sp_configure 'max server memory', 23424; -- Paste your "normal" value here!
RECONFIGURE WITH OVERRIDE;
